// Started with ancient code developed by Matthew Phillips
// It has been HEAVILY modified from his base code but I thought he 
// deserved some credit for getting the ball rolling :)

// For any questions, improvements or suggestions please contact at:
// Kelsey Martins -> kelseymartins@hotmail.com

// Free for any use with exception that you CANNOT sell this code or 
// make profit from it.

// Version 1.0

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;

public class NumericSpinner extends JComponent
{
    protected static final int MIN_ARROW_HEIGHT = 3;
    protected static final int MIN_BUTTON_WIDTH = 17;

    public NumericSpinner()
    {        
        super();        
        bIntegerVersion = true;
        dMinimum = 0;
        dMaximum = 100;
        dIncrement = 1;
        dValue = 0;
        SetPrecision(0);
	init ();
    }

    public NumericSpinner(int minimum, int maximum, int value, int increment)
    {
        super();        
        bIntegerVersion = true;
        dMinimum = (double)(minimum);
        dMaximum = (double)(maximum);
        dIncrement = (double)(increment);
        dValue = (double)(value);
        SetPrecision(0);
        init();
    }
    
    public NumericSpinner(double minimum, double maximum, double value, double increment, int precision)
    {
        super();        
        bIntegerVersion = false;
        dMinimum = minimum;
        dMaximum = maximum;
        dIncrement = increment;
        dValue = value;
        SetPrecision(precision);
        init();
    }

    // Get the current value.
    public int getValue() { return (int)(dValue); }
    public double getValueDouble() { return dValue; }
    public String getValueString()
    {
        if (bIntegerVersion)
        {
            return Integer.toString(getValue());
        }
        else
        {
            return PrecisionFormat.format(getValueDouble());
        }
    }

    // Set the current value
    public void setValue(int value) { dValue = (double)(value); loadTextField(); }
    public void setValue(double value) { dValue = value; loadTextField(); }

    // Get the current minimum value.
    public int getMinimum() { return (int)(dMinimum); }
    public double getMinimumDouble() { return dMinimum; }

    // Set the current minimum value.
    public void setMinimum(int minimum) { dMinimum = (double)(minimum); }
    public void setMinimum(double minimum) { dMinimum = minimum; }
	
    // Get the current maximum value.	 
    public int getMaximum() { return (int)(dMaximum); }
    public double getMaximumDouble() { return dMaximum; }

    // Set the current model.
    public void setMaximum(int maximum) { dMaximum = (double)(maximum); }
    public void setMaximum(double maximum) { dMaximum = maximum; }

    // Returns the increment of the spinner
    public int getIncrement() { return (int)(dIncrement); }
    public double getIncrementDouble() { return dIncrement; }

    // Sets the increment of the spinner
    public void setIncrement(int increment) { dIncrement = (double)(increment); }
    public void setIncrement(double increment) { dIncrement = increment; }

    public void setToolTipText(String text) { textField.setToolTipText(text); }
    public String getActualText()
    {
        return textField.getText();
    }
    
    public boolean getAllowBlank() { return bAllowBlank; }
    public void setAllowBlank(boolean allow) { bAllowBlank = allow; }
    
    public void setEnabled(boolean enable)
    {          
        super.setEnabled(enable);
        textField.setEnabled(enable);
        upButton.setEnabled(enable);
        downButton.setEnabled(enable);
    }

    public void setVisible(boolean visible)
    {                
        super.setVisible(visible);
        textField.setVisible(visible);
        upButton.setVisible(visible);
        downButton.setVisible(visible);        
    }
    
    public void clearValue()
    {
        changing = true;
        if (bIntegerVersion)
        {
            setValue((int)(dMinimum));
        }
        else
        {
            setValue(dMinimum);
        }
        textField.setText("");
        changing = false;
    }
    
    private void SetPrecision(int precision)
    {
        iPrecision = precision;
        switch (precision)
        {
            case 1:
                PrecisionFormat = new DecimalFormat("0.0");
                break;
            case 2:
                PrecisionFormat = new DecimalFormat("0.00");
                break;
            case 3:
                PrecisionFormat = new DecimalFormat("0.000");
                break;
            case 4:
                PrecisionFormat = new DecimalFormat("0.0000");
                break;
            case 5:
                PrecisionFormat = new DecimalFormat("0.00000");
                break;
            default:
                PrecisionFormat = new DecimalFormat("0");
        }        
    }    
    public int getPrecision() { return iPrecision; }
    
    protected void init()
    {
        bAllowBlank = true;
        setLayout(new LayoutManager());
        textField = new NumericTextField();
	loadTextField();
        textField.setFont(new java.awt.Font("Dialog", 1, 10));

	upButton = new ArrowButton();
	downButton = new ArrowButton();

	add(textField);
	add(upButton);
	add(downButton);

	ButtonChangeListener buttonChangeListener = new ButtonChangeListener();
	upButton.getModel().addChangeListener(buttonChangeListener);
	downButton.getModel().addChangeListener(buttonChangeListener);

	// add textField listeners
	TextFieldListener textListener = new TextFieldListener();
	//textField.addActionListener(textListener);
	textField.addFocusListener(textListener);
    }

    protected synchronized void doIncrement(double increment)
    {
        try
        {
            if (bIntegerVersion)
            {
                int i = checkChangeValue((int)(dValue) + (int)(increment));
                if (i == 1)
                {
                    setValue((int)(dMaximum));
                }
                else if (i == -1)
                {
                    setValue((int)(dMinimum));
                }
                else
                {
                    setValue((int)(dValue) + (int)(increment));
                }
            }
            else
            {
                int i = checkChangeValue(dValue + increment);
                if (i == 1)
                {
                    setValue(dMaximum);
                }
                else if (i == -1)
                {
                    setValue(dMinimum);
                }
                else
                {
                    setValue(dValue + increment);
                }
            }
        } 
        catch (IllegalArgumentException exception)
        {}
    }

    // 0 = ok, -1 = min value limit trigger, 1 = max value limit trigger
    protected int checkChangeValue(double val)
    {
        int iRetValue = 0;
        if (val > dMaximum) iRetValue = 1;        
        else if (val < dMinimum) iRetValue = -1;
        return iRetValue;
    }
    
    // 0 = ok, -1 = min value limit trigger, 1 = max value limit trigger
    protected int checkChangeValue(int val)
    {
        int iRetValue = 0;
        if (val > (int)(dMaximum)) iRetValue = 1;        
        else if (val < (int)(dMinimum)) iRetValue = -1;
        return iRetValue;
    }    
    
    // Loads the text field value from the model.  Sets <code>changing</code> 
    // to prevent circular change event propagation.
    protected synchronized void loadTextField()
    {
        if (!changing)
	{
            changing = true;
            if (bIntegerVersion)
            {
                textField.setText(Integer.toString((int)(dValue)));
            }
            else
            {                                
                java.text.DecimalFormat PrecisionFormat = new DecimalFormat("0.00000");
                textField.setText(PrecisionFormat.format(dValue));
            }
            changing = false;
	}
    }


    // Loads the model from the text field.  Sets <code>changing</code>
    // to prevent circular change event propagation.
    protected synchronized void loadWithTextField()
    {
        if (!changing)
        {
            changing = true;
            int i = 0;
            try
            {
                if (bIntegerVersion)
                {
                    int iNewValue = Integer.parseInt(textField.getText());
                    i = checkChangeValue(iNewValue);
                    if (i == 1) 
                    {
                        setValue((int)(dMaximum));
                    }
                    else if (i == -1)
                    {
                        setValue((int)(dMinimum));
                    }
                    else
                    {
                        setValue(iNewValue);
                    }
                }
                else
                {
                    double dNewValue = Double.parseDouble(textField.getText());
                    i = checkChangeValue(dNewValue);
                    if (i == 1) 
                    {
                        setValue(dMaximum);
                    }
                    else if (i == -1)
                    {
                        setValue(dMinimum);
                    }
                    else
                    {
                        setValue(dNewValue);
                    }                    
                }
            } 
            catch (NumberFormatException exception)
            { }
            changing = false;
            if (i != 0)
            {
                loadTextField();
            }
        }
    }

    // Extends JButton to implement arrow buttons.
    protected class ArrowButton extends JButton
    {
        public ArrowButton()
        {
            super();
            init();
	}

	public ArrowButton(Icon icon)
	{
            super(icon);
            init();
	}

        protected void init()
	{
            setFocusPainted(false);
            setMargin(buttonMargin);
	}

	public boolean isFocusTraversable()
	{
            return false;
	}
    }
    
    // Listener for text field
    //protected class TextFieldListener implements ActionListener, FocusListener, Serializable
    protected class TextFieldListener implements FocusListener, Serializable
    {         
        /*public void actionPerformed(ActionEvent event)
	{          
            loadWithTextField();
	}*/

        public void focusGained(FocusEvent event)
	{}

	public void focusLost(FocusEvent event)
	{
            if (bAllowBlank)
            {
                loadWithTextField();
            }
            else
            {
                if (textField.getText().length() == 0)
                {
                    loadTextField();
                }
            }
	}
    }

    // ChangeListener for button models
    // Listens for changes to the arrow button models and timer ticks
    // that signal auto repeats when buttons are held down.
    protected class ButtonChangeListener implements ChangeListener, ActionListener, Serializable
    {
        public ButtonChangeListener()
        {
            autoRepeatTimer = new Timer(100, this);
            autoRepeatTimer.setInitialDelay(300);
        }

        public void stateChanged(ChangeEvent event)
        {
            loadWithTextField();
            ButtonModel buttonModel = (ButtonModel)event.getSource();			
            if (buttonModel.isPressed())
            {
                if (upButton.getModel() == buttonModel)
                {
                    currentIncrement = dIncrement;
                }
                else
                {
                    currentIncrement = -dIncrement;
                }
                doIncrement(currentIncrement);
                autoRepeatTimer.restart();
            } 
            else
            {
                autoRepeatTimer.stop();
            }
        }

        // called by autoRepeatTimer
        public void actionPerformed(ActionEvent event)
        {
            doIncrement(currentIncrement);
        }

        protected Timer autoRepeatTimer;
        protected double currentIncrement;
    }

    // LayoutManager
    // Custom layout manager for NumericSpinner.  Auto-sizes buttons and
    // text field for best effect based on desired height of text field.
    protected class LayoutManager implements java.awt.LayoutManager, Serializable
    {
        public LayoutManager() {}

        public void addLayoutComponent(String name, Component component) {}
		
        public void removeLayoutComponent(Component component) {}

        public Dimension minimumLayoutSize(Container parent)
        {
            Dimension size = textField.getPreferredSize();
            size.height = (MIN_ARROW_HEIGHT + 2) * 2;
            size.width = (MIN_BUTTON_WIDTH + size.width);
            return size;
        }

        public Dimension preferredLayoutSize(Container parent)
        {
            return minimumLayoutSize(parent);
        }

        public void layoutContainer(Container parent)
        {
            // size components
            upButton.setIcon(new ArrowIcon (ArrowIcon.UP, (int)(getSize().height / 2) - 6));
            downButton.setIcon(new ArrowIcon (ArrowIcon.DOWN, (int)(getSize().height / 2) - 6));
            Dimension buttonSize = new java.awt.Dimension(MIN_BUTTON_WIDTH, (int)(getSize().height / 2));
            upButton.setSize(buttonSize);            
            downButton.setSize(buttonSize);

            Dimension textSize = new Dimension(getSize().width - buttonSize.width, getSize().height);
            textField.setSize(textSize);

            // locate components
            Insets insets = getInsets();
            textField.setLocation(insets.left, insets.top);
            int buttonX = textField.getSize().width + insets.left;
            upButton.setLocation(buttonX, insets.top);
            downButton.setLocation(buttonX, upButton.getSize().height + insets.top);
        }

    }

    public class NumericTextField extends JTextField
    {       
        protected void processKeyEvent(KeyEvent e)
        {                   
            char c = e.getKeyChar();
            int i = e.getKeyCode();
            if (('0' <= c && c <= '9') || ('\b' == c) || ('\n' == c) ||
                (i == e.VK_BACK_SPACE) || (i == e.VK_DELETE) || (i == e.VK_TAB) ||
                (i == e.VK_LEFT) || (i == e.VK_RIGHT) || 
                ((bIntegerVersion == false) && ('.' == c) && (this.getText().indexOf('.') < 0)))
            {                                
                super.processKeyEvent(e);
                if ((this.getText().length() > 0) && ((i != e.VK_LEFT) && (i != e.VK_RIGHT)))
                {
                    loadWithTextField();
                }
                else if (this.getText().length() == 0)
                {
                    changing = true;
                    if (bIntegerVersion)
                    {                        
                        setValue((int)(dMinimum));                        
                    }
                    else
                    {
                        setValue(dMinimum);
                    }
                    changing = false;
                }
            }
        }
    }
    
    protected static final Insets buttonMargin = new Insets(1, 1, 1, 1);
    
    protected NumericTextField textField;
    protected ArrowButton upButton, downButton;
    protected transient boolean changing = false;
    
    private boolean bAllowBlank;
    private boolean bIntegerVersion;
    private int iPrecision;
    private double dMinimum;
    private double dMaximum;
    private double dIncrement;
    private double dValue;
    private java.text.DecimalFormat PrecisionFormat;
}
